#include <stdio.h>
#include <gb_graph.h>
#include <gb_save.h>

#include "exp.h"
#include "graph.h"
#include "hop.h"
#include "nodeid.h"
#include "pick.h"
#include "route.h"
#include "stress.h"

/************************************************/
/* initialize output file names			*/
/************************************************/

void init_output_file_names(char *topology_fn, char *out_dir,
	char *hop_outfile1, char *hop_outfile2,
	char *rdp_outfile, char *stress_outfile1, char *stress_outfile2,
	int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
	int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
	int group_size, int vnn) {

  char param[MAX_FILENAME_LEN];

  //  int vnn = atoi(getenv("REXEC_MY_VNN"));

  // initialize output file names
  sprintf(param, "%s/%s.%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE, group_size, vnn);
  sprintf(hop_outfile1, "%s.hop.sp", param);
  sprintf(hop_outfile2, "%s.hop.tapestry", param);
  sprintf(rdp_outfile, "%s.hop.rdp", param);
  sprintf(stress_outfile1, "%s.stress.unicast", param);
  sprintf(stress_outfile2, "%s.stress.tapestry", param);

  printf("%s\n", hop_outfile1); printf("%s\n", hop_outfile2);
  printf("%s\n", rdp_outfile);
  printf("%s\n", stress_outfile1); printf("%s\n", stress_outfile2);
}

/************************************************/
/* initialize experiment                        */
/* 1.  initialize input topology file name	*/
/* 2.  restore graph				*/
/* 3.  reset arc lengths to 1			*/
/* 4.  clear picked_t field of every vertex	*/
/* 5.  assign tapestry node IDs to vertices	*/
/* 6.  build tapestry routing table		*/
/************************************************/
  
Graph *exp_init(char *topology_fn, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE) {
  
  Graph *g;
  char gb_infile[MAX_FILENAME_LEN]; 
                                
  // initialize input file name
  sprintf(gb_infile, "../data/%s", topology_fn);

  g = restore_graph(gb_infile); // restores graph g from a sgb data file

  if (g == NULL)
    printf("ERROR: graph not read from topology data file.\n");
  reset_arc_len(g);             // reset arc lengths to 1

  clear_picked_t_field(g);	// clear the picked_t field of every vertex

  // assign tapestry nodeIDs to g's vertices
  assign_tapestry_nodeID(g, TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE,
                                TAPESTRY_NAMESPACE_SIZE);
  
  //  print_all_arcs(g);
 
  // build tapestry routing table
  build_tapestry_routing_table(g, TAPESTRY_NODEID_SIZE,
                                TAPESTRY_NAMESPACE_BASE);

  // print tapestry routing table
  //  print_tapestry_routing_table(g, TAPESTRY_NODEID_SIZE,
  //                                TAPESTRY_NAMESPACE_BASE);
  return g;
}

/************************************************/
/* run experiment				*/
/* 1.  initialize output file names		*/
/* 2.  clear picked_m field of every vertex	*/
/* 3.  pick a random sender			*/
/* 4.  pick random multicast receivers		*/
/* 5.  calc # of hops from sender to each	*/
/*     receiver via shortest path routing	*/
/* 6.  calc # of hops from sender to each	*/
/*     receiver via tapestry routing		*/
/* 7.  compute RDP				*/
/* 8.  calc stress in unicast routing		*/
/* 9.  calc stress in tapestry routing		*/
/************************************************/

void exp_run(Graph *g, char *topology_fn, char *out_dir,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
                int group_size, int vnn) {

  char hop_outfile1[MAX_FILENAME_LEN];
  char hop_outfile2[MAX_FILENAME_LEN];
  char rdp_outfile[MAX_FILENAME_LEN];
  char stress_outfile1[MAX_FILENAME_LEN];
  char stress_outfile2[MAX_FILENAME_LEN];

  Vertex *sender;
  Vertex **recvr_array;         
  long *hop_array1, *hop_array2;

  printf("\ntopology size = %ld, tapestry size = %d, group size = %d\n",
                g->n, TAPESTRY_NAMESPACE_SIZE, group_size);

  init_output_file_names(topology_fn, out_dir,
		hop_outfile1, hop_outfile2, rdp_outfile,
		stress_outfile1, stress_outfile2, 
		TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
		TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
		group_size, vnn);

  clear_picked_m_field(g);

  // pick a random sender
  sender = pick_1(g);

  // pick random multicast receivers from the tapestry network
  recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
  pick_n(g, recvr_array, group_size - 1);

  // calc # of hops from the sender to each member via shortest path
  hop_calc_sp(g, sender, recvr_array, group_size - 1, hop_outfile1);

  // calc # of hops from the sender to each member via tapestry routing
  hop_calc_tapestry(g, sender, recvr_array, group_size - 1, hop_outfile2,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);

  // compute RDP = tapestry routing delay / shortest path delay
  hop_array1 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_array2 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_read(hop_array1, group_size - 1, hop_outfile1);
  hop_read(hop_array2, group_size - 1, hop_outfile2);
  rdp_calc(hop_array1, hop_array2, group_size - 1, rdp_outfile);

  // compute stress of a physical link = # of identifical copies of
  // a packet carried by a physical link
  clear_stress_field(g);
  stress_calc_unicast(g, sender, recvr_array, group_size-1, stress_outfile1);
  clear_stress_field(g);
  stress_calc_tapestry(g, sender, recvr_array, group_size-1, stress_outfile2,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
}
