#include <string.h>
#include <math.h>
#include <gb_graph.h>
#include <gb_dijk.h>

#include "exp.h"
#include "graph.h"
#include "hop.h"
#include "nodeid.h"
#include "stack.h"
#include "util.h"

/**************
 ** calculate IP distance in locating objects
 **/
int hop_calc_loc_IP(Graph *g, Vertex *client, Vertex *server) {
  dijkstra(client, server, g, NULL);
  return server->dist;
}

/************************************************/
/* calculate shortest path distances from       */
/* the sender to each receiver in the           */
/* recvr_array and write to hop_outfile         */
/************************************************/

void hop_calc_sp(Graph *g, Vertex *sender, Vertex **recvr_array, int n,
                                                        char *hop_outfile) {
  int i, j;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v = NULL;

  printf("\nhop_calc_sp:  sender %s (%s)\n", sender->name, sender->nodeid);
  fflush(stdout);
  for (i = 0; i < n; i++) {
    dijkstra(sender, recvr_array[i], g, NULL);
    hop_array[i] = (recvr_array[i])->dist;
    printf("%s(%ld): ", (recvr_array[i])->nodeid, hop_array[i]);
    fflush(stdout);

    v = recvr_array[i];
    for (j = 0; j < hop_array[i]; j++) {
      printf("%s ", v->backlink->nodeid); fflush(stdout);
      v = v->backlink;
    }
    printf("\n"); fflush(stdout);
  }

  // write shortest paths to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
} 
  
/************************************************/
/* calculate number of hops in tapestry routing */
/* from the sender to each receiver in the      */
/* recvr_array and write to the hop_outfile     */
/************************************************/
   
void hop_calc_tapestry(Graph *g, Vertex *sender, Vertex **recvr_array, int n,
                char *hop_outfile, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  int i, j, distance, mask, digit, r_id;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v1, *v2, *r;
  Vertex **routing_table;

  printf("\nhop_calc_tapestry:  sender %s (%s)\n",
					sender->name, sender->nodeid);
  fflush(stdout);

  for (i = 0; i < n; i++) {
    v1 = sender;
    r = recvr_array[i];
    distance = 0;
    for (j = 0; j < TAPESTRY_NODEID_SIZE; j++) {
      routing_table = (Vertex **) v1->rtable;
      mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * j);
      tapestry_nodeID_atoi(r->nodeid, &r_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * j);
      v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j + digit];
      dijkstra(v1, v2, g, NULL);
      printf("%s(%ld) ", v2->nodeid, v2->dist); fflush(stdout);
      distance += v2->dist;
      v1 = v2;
    }
    hop_array[i] = distance;
    printf(" %s(%ld)\n", (recvr_array[i])->nodeid, hop_array[i]);
    fflush(stdout);
  }
  printf("\n"); fflush(stdout);

  // write shortest paths to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/* read n shortest path distances from a file   */
/************************************************/

void hop_read(long *hop_array, int n, char *hop_outfile) {
  FILE *hop_fp = fopen(hop_outfile, "r");

  fread(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/* read n neighbors from a file			*/
/************************************************/

void hop_read_maxconn(long *nbr_array, int n, char *hop_outfile) {
  FILE *hop_fp = fopen(hop_outfile, "r");

  fseek(hop_fp, sizeof(long) * n, SEEK_SET);
  fread(nbr_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/* calculate RDP between two members along the  */
/* tapestry overlay to the unicast delay        */
/* between them                                 */
/************************************************/

void rdp_calc(long *hop_array1, long *hop_array2, int n, char *rdp_outfile) {
  int i;
  FILE *rdp_fp = fopen(rdp_outfile, "w");
  float *rdp_array = (float *) malloc(sizeof(float) * n);

  printf("rdp_calc:\n"); fflush(stdout);
  for (i = 0; i < n; i++) {
    rdp_array[i] = (float) (hop_array2[i] * 1.0 / hop_array1[i]);
    printf("%.2f ", rdp_array[i]); fflush(stdout);
  }
  printf("\n"); fflush(stdout);

  // write relative delay penalties to file in binary
  fwrite(rdp_array, sizeof(float), n, rdp_fp);
  fclose(rdp_fp);
}

/************************************************/
/* calculate shortest path distances from       */
/* the sender to each receiver in the           */
/* recvr_array and write to hop_outfile         */
/************************************************/

void hop_calc_sp_maxconn(Graph *g, Vertex *sender, Vertex **recvr_array,
        Graph *new_graph, Vertex *new_sender, Vertex **new_recvr_array,
        int n, char *hop_outfile) {
  int i, j;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  long distance1, distance2;
  Vertex *v, *v1, *v2, *u1, *u2;

  printf("\nhop_calc_sp_maxconn:  sender %s (%s)\n", new_sender->name,
                                                new_sender->nodeid);
  fflush(stdout);

  if (strcmp(sender->name, new_sender->name) != 0) {
    printf("ERROR:  sender %s, new_sender %s\n",
                                sender->name, new_sender->name);
    fflush(stdout);
    exit(1);
  }
  for (i = 0; i < n; i++) {
    if (strcmp((recvr_array[i])->name, (new_recvr_array[i])->name) != 0) {
      printf("ERROR:  recvr_array[%d] %s, new_recvr_array[%d] %s\n",
                i, (recvr_array[i])->name, i, (new_recvr_array[i])->name);
      fflush(stdout);
      exit(1);
    } 
  }

  for (i = 0; i < n; i++) {
    for (v = g->vertices; v < g->vertices + g->n; v++) {
      v->backlink = NULL;
    }
    for (v = new_graph->vertices; v < new_graph->vertices+new_graph->n; v++) {
      v->backlink = NULL;
    }
    distance1 = dijkstra(sender, recvr_array[i], g, NULL);
    distance2 = dijkstra(new_sender, new_recvr_array[i], new_graph, NULL);

    if (distance2 == -1)        // if there's no longer a path from s->r
      hop_array[i] = -1;
    else {                      // else there's still a path from s->r
      v1 = recvr_array[i];
      v2 = new_recvr_array[i];
      for (j = 0; j < distance1; j++) {
        u1 = v1->backlink;
        u2 = v2->backlink;
        if (u2 == NULL)
          break;
        if (strcmp(u1->name, u2->name) != 0)
          break;
        v1 = u1;
        v2 = u2;
      } 
      if (j == distance1)       // if shortest paths match in g and new_graph
        hop_array[i] = distance1;
      else                      // else original path no longer exists
        hop_array[i] = 0;
    }
    printf("hop_array[%d] = %ld\n", i, hop_array[i]); fflush(stdout);
    fflush(stdout);
  }

  // write shortest paths to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
} 

/************************************************/
/* calculate number of hops in tapestry routing */
/* from the sender to each receiver in the      */
/* recvr_array and write to the hop_outfile     */
/************************************************/
   
void hop_calc_tapestry_maxconn(Graph *new_graph, Vertex *new_sender,
                Vertex **new_recvr_array, int n,
                char *hop_outfile, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  int i, j, k, distance, mask, digit, r_id, dijk_res;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v1, *v2, *r;
  Vertex **routing_table;
  stack *stk = NULL;
  stack_elem *new_elem, *top_elem;

  printf("\nhop_calc_tapestry_maxconn:  sender %s (%s)\n",
                                new_sender->name, new_sender->nodeid);
  fflush(stdout);

  for (i = 0; i < n; i++) {
    v1 = new_sender;
    r = new_recvr_array[i];

    distance = dijkstra(new_sender, new_recvr_array[i], new_graph, NULL);

    if (distance == -1) {
      hop_array[i] = -1;
    }
    else {
      stk = (stack *) malloc(sizeof(stack));
      stack_initialize(stk);

      distance = 0;
      k = 0;
      j = 0;
      while (j < TAPESTRY_NODEID_SIZE) {
        routing_table = (Vertex **) v1->rtable;
        mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * j);
        tapestry_nodeID_atoi(r->nodeid, &r_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
        digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * j);

        while (k < MAX_NUM_NBRS) {
          v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j * MAX_NUM_NBRS +
                                digit * MAX_NUM_NBRS + k];
          dijk_res = dijkstra(v1, v2, new_graph, NULL);
          if (dijk_res != -1) {
            new_elem = create_elem(v1, k+1, j, distance);
            stack_push(stk, new_elem);
            distance += v2->dist;
            break;
          }
          k++;
        }
        if (dijk_res == -1) {
          while (!stack_empty(stk)) {
            top_elem = stack_top(stk);
            if (top_elem->next_nbr < MAX_NUM_NBRS) {
              v1 = top_elem->current_v;
              k = top_elem->next_nbr;
              j = top_elem->level;
              distance = top_elem->distance;
              break;
            }
            else {
              stack_pop(stk);
            }
          }
          if (stack_empty(stk)) {
            distance = -1;
            break;
          }
        }
        else {
          v1 = v2;
          k = 0;
          j++;
        }
      }
      hop_array[i] = distance;
      stack_free(stk);
    }
    printf(" %s(%ld)\n", (new_recvr_array[i])->nodeid, hop_array[i]);
    fflush(stdout);
  }

  // write shortest paths and neighbors to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/************************************************/

void hop_calc_sp_partition(Graph *g, Vertex **recvr_array,
                int n, char *hop_outfile) {
  int i, j;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v = NULL, *sender = NULL;

  printf("\nhop_calc_sp_partition\n");
  fflush(stdout);
  for (i = 0; i < n; i++) {
    sender = (recvr_array[i])->mroot;
    dijkstra(sender, recvr_array[i], g, NULL);
    hop_array[i] = (recvr_array[i])->dist;
    printf("%s(%ld): ", (recvr_array[i])->nodeid, hop_array[i]);
    fflush(stdout);

    v = recvr_array[i];
    for (j = 0; j < hop_array[i]; j++) {
      printf("%s ", v->backlink->nodeid); fflush(stdout);
      v = v->backlink;
    }
    printf("\n"); fflush(stdout);
  }

  // write shortest paths to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/************************************************/

void hop_calc_tapestry_partition(Graph *g, Vertex **recvr_array,
                int n, char *hop_outfile,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
                int TAPESTRY_NAMESPACE_BASE) {
  int i, j, distance, mask, digit, r_id;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v1, *v2, *r, *sender;
  Vertex **routing_table;

  printf("\nhop_calc_tapestry_partition\n"); fflush(stdout);

  for (i = 0; i < n; i++) {
    sender = (recvr_array[i])->mroot;
    v1 = sender;
    r = recvr_array[i];
    distance = 0;
    for (j = 0; j < TAPESTRY_NODEID_SIZE; j++) {
      routing_table = (Vertex **) v1->rtable;
      mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * j);
      tapestry_nodeID_atoi(r->nodeid, &r_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * j);
      v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j + digit];
      dijkstra(v1, v2, g, NULL);
      printf("%s(%ld) ", v2->nodeid, v2->dist); fflush(stdout);
      distance += v2->dist;
      v1 = v2;
    }
    hop_array[i] = distance;
    printf(" %s(%ld)\n", (recvr_array[i])->nodeid, hop_array[i]);
    fflush(stdout);
  }
  printf("\n"); fflush(stdout);

  // write shortest paths to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/************************************************/

void hop_calc_tapestry_protocol(Graph *new_graph, Vertex *new_sender,
                Vertex **new_recvr_array, int n,
                char *hop_outfile, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  int i, j, k, distance, mask, digit, r_id, dijk_res;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v1_p, *v2_p, *v1_s, *v2_s, *r;
  Vertex **routing_table_p, **routing_table_s;

  printf("\nhop_calc_tapestry_protocol:  sender %s (%s)\n",
                                new_sender->name, new_sender->nodeid);
  fflush(stdout);

  for (i = 0; i < n; i++) {
    v1_p = v1_s = new_sender;
    r = new_recvr_array[i];

    distance = dijkstra(new_sender, new_recvr_array[i], new_graph, NULL);

    if (distance == -1) {
      hop_array[i] = -1;
    }
    else {
      distance = 0;
      for (j = 0; j < TAPESTRY_NODEID_SIZE; j++) {
        mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * j);
        tapestry_nodeID_atoi(r->nodeid, &r_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
        digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * j);

        routing_table_p = (Vertex **) v1_p->rtable;
        v2_p = routing_table_p[TAPESTRY_NAMESPACE_BASE * j * MAX_NUM_NBRS +
                                digit * MAX_NUM_NBRS + 0];

        routing_table_s = (Vertex **) v1_s->rtable;
        for (k = 0; k < MAX_NUM_NBRS; k++) {
          v2_s = routing_table_s[TAPESTRY_NAMESPACE_BASE * j * MAX_NUM_NBRS +
                                digit * MAX_NUM_NBRS + k];
          dijk_res = dijkstra(v1_s, v2_s, new_graph, NULL);
          if (dijk_res != -1) {
            // if v2_s is not on the primary path, distribute multicast state
            if (v2_p != v2_s) { 
              distance += ceil(n * 1.0 / pow(TAPESTRY_NAMESPACE_BASE, j+1) / 
                                        NUM_MEMBER_PER_PKT);
            }
            distance += v2_s->dist;
            break;
          }
        }
        if (dijk_res != -1) {
          v1_p = v2_p;
          v1_s = v2_s;
        }
        else {
          distance = -1;
          break;
        }
      }
      hop_array[i] = distance;
    }
    printf(" %s(%ld)\n", (new_recvr_array[i])->nodeid, hop_array[i]);
    fflush(stdout);
  }

  // write shortest paths and neighbors to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}
