#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_basic.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "graph.h"
#include "exp.h"
#include "fault.h"
#include "hop.h"
#include "nodeid.h"
#include "pick.h"
#include "route.h"
#include "stat.h"
#include "stress.h"
#include "util.h"

/************************************************/
/* initialize output file names                 */
/************************************************/

void exp16_init_output_file_names(char *topology_fn, char *out_dir,
        char *hop_outfile1, char *hop_outfile2,
        char *rdp_outfile, char *stress_outfile1, char *stress_outfile2,
        int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
        int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
        int group_size, int percentage, int vnn) {

  char param[MAX_FILENAME_LEN];

  //  int vnn = atoi(getenv("REXEC_MY_VNN"));

  // initialize output file names
  sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, percentage, vnn);
  sprintf(hop_outfile1, "%s.hop.sp", param);
  sprintf(hop_outfile2, "%s.hop.tapestry", param);
  sprintf(rdp_outfile, "%s.hop.rdp", param);
  sprintf(stress_outfile1, "%s.stress.unicast", param);
  sprintf(stress_outfile2, "%s.stress.tapestry", param);

  printf("%s\n", hop_outfile1); printf("%s\n", hop_outfile2);
  printf("%s\n", rdp_outfile);
  printf("%s\n", stress_outfile1); printf("%s\n", stress_outfile2);
}

/************************************************/
/* initialize experiment                        */
/* 1.  assign tapestry node IDs to vertices     */
/* 2.  build tapestry routing table             */
/************************************************/
    
void exp16_init(Graph *g, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE){

  clear_picked_t_field(g);      // clear the picked_t field of every vertex
   
  // assign tapestry nodeIDs to g's vertices
  assign_tapestry_nodeID(g, TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE,
                                TAPESTRY_NAMESPACE_SIZE);
 
  print_all_arcs(g);

  // build tapestry routing table
  build_tapestry_routing_table_maxconn(g, TAPESTRY_NODEID_SIZE,
                                TAPESTRY_NAMESPACE_BASE);
  // print tapestry routing table
  print_tapestry_routing_table_maxconn(g, TAPESTRY_NODEID_SIZE,
                                TAPESTRY_NAMESPACE_BASE);
}     

/************************************************/
/* run experiment                               */
/* 1.  initialize output file names             */
/* 2.  calc # of hops from sender to each       */
/*     receiver via shortest path routing       */
/* 3.  calc # of hops from sender to each       */
/*     receiver via tapestry routing            */
/* 4.  compute RDP                              */
/************************************************/
        
void exp16_run(Graph *g, Vertex *sender, Vertex **recvr_array,
	Graph *new_graph, Vertex *new_sender, Vertex **new_recvr_array,
	char *topology_fn, char *out_dir,
        int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
        int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
        int group_size, int percentage, int vnn) {

  char hop_outfile1[MAX_FILENAME_LEN];
  char hop_outfile2[MAX_FILENAME_LEN];
  char rdp_outfile[MAX_FILENAME_LEN]; 
  char stress_outfile1[MAX_FILENAME_LEN];
  char stress_outfile2[MAX_FILENAME_LEN];
  long *hop_array1, *hop_array2;
  
  printf("\ntopology size = %ld, tapestry size = %d, group size = %d, ",
                g->n, TAPESTRY_NAMESPACE_SIZE, group_size);
  fflush(stdout);

  exp16_init_output_file_names(topology_fn, out_dir,
                hop_outfile1, hop_outfile2, rdp_outfile,
                stress_outfile1, stress_outfile2,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage, vnn);
                                
  // calc # of hops from the sender to each member via shortest path
  // = -1 if original shortest path no longer exists in new_graph
  hop_calc_sp_maxconn(g, sender, recvr_array,
		new_graph, new_sender, new_recvr_array,
		group_size-1, hop_outfile1);

  // calc # of hops from the sender to each member via tapestry routing
  hop_calc_tapestry_protocol(new_graph, new_sender, new_recvr_array,
	group_size-1, hop_outfile2, TAPESTRY_NODEID_SIZE,
	TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);

  // compute RDP = tapestry routing delay / shortest path delay
  hop_array1 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_array2 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_read_maxconn(hop_array1, group_size - 1, hop_outfile1);
  hop_read_maxconn(hop_array2, group_size - 1, hop_outfile2);
  rdp_calc(hop_array1, hop_array2, group_size - 1, rdp_outfile);
}

/************************************************/
/* experiment 16                                */
/************************************************/

void exp_16() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out16";
  char gb_infile[MAX_FILENAME_LEN];
  int group_size = 4096;
  int vnn = 9, i, j, k;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
        TAPESTRY_NAMESPACE_SIZE;
  Graph *g, *new_graph;
  Vertex *sender, **recvr_array, *new_sender, **new_recvr_array;
  int num_links, num_failed_links;
  double percentage = 1;		// 1%

  TAPESTRY_NODEID_SIZE = 4;
  TAPESTRY_DIGIT_SIZE = 3;
  TAPESTRY_NAMESPACE_BASE = 8;          // octal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;

/*
  TAPESTRY_NODEID_SIZE = 6; 
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;          // quad digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
*/
    
  for (i = 0; i < 4; i++) {
    sprintf(gb_infile, "../data/%s", topology_fn_array[i]);
    g = restore_graph(gb_infile);	// restores graph g
    if (g == NULL)
      printf("ERROR: graph not read from topology data file.\n");

    reset_arc_len(g);			// reset arc lengths to 1

    new_graph = restore_graph(gb_infile);
    num_links = reset_arc_len(new_graph);	// reset arc lengths to 1

    exp16_init(new_graph, TAPESTRY_NODEID_SIZE,
             TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

    clear_picked_m_field(new_graph);

    // pick a random sender
    new_sender = pick_1(new_graph);
    sender = g->vertices + (new_sender - new_graph->vertices);

    // pick random multicast receivers from the tapestry network
    new_recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
    pick_n(new_graph, new_recvr_array, group_size - 1);
    recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
    for (j = 0; j < group_size - 1; j++) {
      recvr_array[j] = g->vertices +
				(new_recvr_array[j] - new_graph->vertices);
    }

    k = 0;
    exp16_run(g, sender, recvr_array,
		new_graph, new_sender, new_recvr_array,
		topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage * k, vnn);

    num_failed_links = num_links * percentage / 100;
    for (k = 1; k <= 50; k++) {

      inject_link_faults(new_graph, num_failed_links);

      exp16_run(g, sender, recvr_array,
		new_graph, new_sender, new_recvr_array,
		topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage * k, vnn);
    }
  }
}

/************************************************/
/* MAIN                                         */
/************************************************/ 
    
int main() {
   
  gb_init_rand(time(NULL));     // initialize the random number generator
  
  // experiment for 16, fault tolerant protocol, forward packets to
  // first reachable neighbor (no backtracking)
  // only nodes on the primary path has membership state initially
  exp_16();

  return 0;
}               
