#include <limits.h>
#include <string.h>
#include <math.h>
#include <gb_graph.h>
#include <gb_dijk.h>

#include "exp.h"
#include "graph.h"
#include "nodeid.h"
#include "route.h"

/************************************************/
/* finds the TAPESTRY_NAMESPACE_BASE closest    */
/* neighbors whose node-IDs match the lowest    */
/* (level-1) digits of v's node-ID and who      */
/* have different combinations of the (level)th */
/* digit; one of these is always a loopback     */
/* pointer to node v itself                     */
/************************************************/

void nth_level_closest_neighbors(Graph *g, Vertex *v, int level,
                Vertex **routing_table,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {
  
  int i, v_mask, u_mask, v_suffix, u_suffix, distance, v_id, u_id;
  Vertex *u, *closest_u = NULL;
  char node_id[TAPESTRY_NODEID_SIZE + 1];

  if (level > 0 && level < TAPESTRY_NODEID_SIZE) {
    dijkstra(v, NULL, g, NULL);
    v_mask = pow(TAPESTRY_NAMESPACE_BASE, level-1) - 1;
    u_mask = pow(TAPESTRY_NAMESPACE_BASE, level) - 1;
    tapestry_nodeID_atoi(v->nodeid, &v_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    for (i = 0; i < TAPESTRY_NAMESPACE_BASE; i++) {
      v_suffix = (v_id & v_mask) + ((v_mask + 1) * i);
      distance = INT_MAX; closest_u = NULL;
      for (u = g->vertices; u < g->vertices + g->n; u++) {
        if (u->picked_t == 1) {
          tapestry_nodeID_atoi(u->nodeid, &u_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
          u_suffix = u_id & u_mask;
          if (u_suffix == v_suffix && u->dist < distance) {
            distance = u->dist;
            closest_u = u;
          }
        }
      }
      routing_table[(level-1) * TAPESTRY_NAMESPACE_BASE + i] = closest_u;
    }
  }
  else if (level == TAPESTRY_NODEID_SIZE) {
    v_mask = pow(TAPESTRY_NAMESPACE_BASE, level-1) - 1;
    tapestry_nodeID_atoi(v->nodeid, &v_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    v_suffix = v_id & v_mask;
    for (i = 0; i < TAPESTRY_NAMESPACE_BASE; i++) {
      tapestry_nodeID_itoa((v_mask + 1) * i + v_suffix, node_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      for (u = g->vertices; u < g->vertices + g->n; u++) {
        if (u->picked_t == 1 && (strcmp(u->nodeid, node_id) == 0))
          closest_u = u;
      }         
      if (closest_u == NULL) {
        printf("ERROR:  lookup of node ID %s failed!\n", node_id);
        exit(1);
      }
      else { 
        routing_table[(level-1) * TAPESTRY_NAMESPACE_BASE + i] = closest_u;
      }
      closest_u = NULL;
    }
  } 
}   
    
/************************************************/
/* build routing table for each tapestry node,  */ 
/* where each table contains                    */ 
/* TAPESTRY_NODEID_SIZE x                       */
/* TAPESTRY_NAMESPACE_BASE entries              */
/* = (3 digits) x 16 = 48 (hexadecimal)         */
/* = (4 digits) x 8 = 32 (octal)                */
/************************************************/
          
void build_tapestry_routing_table(Graph *g,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {
            
  Vertex *v, **routing_table;
  int i, j;
  int routing_table_size = TAPESTRY_NODEID_SIZE * TAPESTRY_NAMESPACE_BASE;

  printf("building tapestry routing table:\n");

  for (v = g->vertices, j = 0; v < g->vertices + g->n; v++, j++) {
    if (v->picked_t == 1) {     // if v is a tapestry node
      routing_table = (Vertex **) malloc(sizeof(Vertex*)*routing_table_size);
      for (i = 0; i < TAPESTRY_NODEID_SIZE; i++) {
        nth_level_closest_neighbors(g, v, i+1, routing_table,
                TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      }
      //      printf("%d(%s) ", j, v->nodeid);
      v->rtable = (long) routing_table;
    }
  }
}

/************************************************/
/* print tapestry routing table of each         */
/* tapestry node                                */
/************************************************/

void print_tapestry_routing_table(Graph *g,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  Vertex *v, *u, **rt;
  int routing_table_size = TAPESTRY_NODEID_SIZE * TAPESTRY_NAMESPACE_BASE;
  int i;

  printf("\n\ntapestry routing table:\n");
  for (v = g->vertices; v < g->vertices + g->n; v++) {
    if (v->picked_t == 1) {
      printf("%s: ", v->nodeid);
      rt = (Vertex **) v->rtable;
      for (i = 0; i < routing_table_size; i++) {
        u = rt[i];
        printf("%s ", u->nodeid);
      }
      printf("\n");
    }
  }
}

/************************************************/
/* finds the TAPESTRY_NAMESPACE_BASE closest    */
/* neighbors whose node-IDs match the lowest    */
/* (level-1) digits of v's node-ID and who      */
/* have different combinations of the (level)th */
/* digit; one of these is always a loopback     */
/* pointer to node v itself                     */
/************************************************/

void nth_level_closest_neighbors_maxconn(Graph *g, Vertex *v, int level,
                Vertex **routing_table,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {  
                        
  int i, j, k, v_mask, u_mask, v_suffix, u_suffix, distance, v_id, u_id;
  Vertex *u, *w, *closest_u = NULL;
  char node_id[TAPESTRY_NODEID_SIZE + 1];
  
  if (level > 0 && level < TAPESTRY_NODEID_SIZE) { 
    dijkstra(v, NULL, g, NULL);
    v_mask = pow(TAPESTRY_NAMESPACE_BASE, level-1) - 1;
    u_mask = pow(TAPESTRY_NAMESPACE_BASE, level) - 1;
    tapestry_nodeID_atoi(v->nodeid, &v_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
                        
    for (i = 0; i < TAPESTRY_NAMESPACE_BASE; i++) { 
      v_suffix = (v_id & v_mask) + ((v_mask + 1) * i);
          
      // find MAX_NUM_NBRS closest neighbors
      for (j = 0; j < MAX_NUM_NBRS; j++) {
        distance = INT_MAX; closest_u = NULL;
        for (u = g->vertices; u < g->vertices + g->n; u++) {
          if (u->picked_t == 1) {
            tapestry_nodeID_atoi(u->nodeid, &u_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
            u_suffix = u_id & u_mask;
            if (u_suffix == v_suffix && u->dist < distance) {
              for (k = 0; k < j; k++) {
                w = routing_table[(level-1) * TAPESTRY_NAMESPACE_BASE *
                        MAX_NUM_NBRS + i * MAX_NUM_NBRS + k];
                if (u == w)
                  break;
              }
              if (j == k) {
                distance = u->dist;
                closest_u = u;                  
              }
            }
          }
        }       
        routing_table[(level-1) * TAPESTRY_NAMESPACE_BASE * MAX_NUM_NBRS +
                        i * MAX_NUM_NBRS + j] = closest_u;
      }
    }
  }
  else if (level == TAPESTRY_NODEID_SIZE) {
    v_mask = pow(TAPESTRY_NAMESPACE_BASE, level-1) - 1;
    tapestry_nodeID_atoi(v->nodeid, &v_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    v_suffix = v_id & v_mask;
    for (i = 0; i < TAPESTRY_NAMESPACE_BASE; i++) {
      tapestry_nodeID_itoa((v_mask + 1) * i + v_suffix, node_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      for (u = g->vertices; u < g->vertices + g->n; u++) {
        if (u->picked_t == 1 && (strcmp(u->nodeid, node_id) == 0)) {
          closest_u = u;
          break;
        } 
      } 
      if (closest_u == NULL) {
        printf("ERROR:  lookup of node ID %s failed!\n", node_id);
        exit(1);
      }                 
      else {
        for (j = 0; j < MAX_NUM_NBRS; j++) {
          routing_table[(level-1) * TAPESTRY_NAMESPACE_BASE * MAX_NUM_NBRS +
                        i * MAX_NUM_NBRS + j] = closest_u;
        }               
      }         
      closest_u = NULL; 
    }         
  }           
}               

/************************************************/
/* build routing table for each tapestry node,  */ 
/* where each table contains                    */ 
/* TAPESTRY_NODEID_SIZE x                       */
/* TAPESTRY_NAMESPACE_BASE entries              */
/* = (3 digits) x 16 = 48 (hexadecimal)         */
/* = (4 digits) x 8 = 32 (octal)                */
/************************************************/
          
void build_tapestry_routing_table_maxconn(Graph *g,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {
            
  Vertex *v, **routing_table;
  int i, j;
  int routing_table_size = TAPESTRY_NODEID_SIZE * TAPESTRY_NAMESPACE_BASE *
                                MAX_NUM_NBRS;

  printf("building tapestry routing table:\n");

  for (v = g->vertices, j = 0; v < g->vertices + g->n; v++, j++) {
    if (v->picked_t == 1) {     // if v is a tapestry node
      routing_table = (Vertex **) malloc(sizeof(Vertex*)*routing_table_size);
      for (i = 0; i < TAPESTRY_NODEID_SIZE; i++) {
        nth_level_closest_neighbors_maxconn(g, v, i+1, routing_table,
                TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      }
      //      printf("%d(%s) ", j, v->nodeid);
      v->rtable = (long) routing_table;
    }
  }
}

/************************************************/
/* print tapestry routing table of each         */
/* tapestry node                                */
/************************************************/

void print_tapestry_routing_table_maxconn(Graph *g,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  Vertex *v, *u, **rt;
  int routing_table_size = TAPESTRY_NODEID_SIZE * TAPESTRY_NAMESPACE_BASE *
                                MAX_NUM_NBRS;
  int i;

  printf("\n\ntapestry routing table:\n");
  for (v = g->vertices; v < g->vertices + g->n; v++) {
    if (v->picked_t == 1) {
      printf("%s: ", v->nodeid);
      rt = (Vertex **) v->rtable;
      for (i = 0; i < routing_table_size; i++) {
        u = rt[i];
        printf("%s ", u->nodeid);
      }
      printf("\n");
    }
  }
}
