#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_basic.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "graph.h"
#include "exp.h"
#include "hop.h"
#include "nodeid.h"
#include "pick.h"
#include "route.h"
#include "stat.h"
#include "stress.h"
#include "util.h"

/************************************************/
/* clear the mroot field of each vertex of	*/
/* the Graph pointed to by g                    */
/************************************************/

void clear_mroot_field(Graph *g) {
  Vertex *v;

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    v->mroot = NULL;
  }
}

/************************************************/
/* clear the locptr field of each vertex of	*/
/* the Graph pointed to by g                    */
/************************************************/

void clear_locptr_field(Graph *g) {
  Vertex *v;

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    v->locptr = 0;
  }
}

/************************************************/
/************************************************/

void insert_object(Graph *g, Vertex *loc_root, Vertex *sender,
		int num_roots, int TAPESTRY_NODEID_SIZE,
		int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  int i, mask, mroot_mask, digit, r_id;
  Vertex *v1, *v2, *r;
  Vertex **routing_table;

  // publish the object stored at sender by marking
  // the nodes from sender up to the loc_root
  mroot_mask = 1 << num_roots;
  sender->locptr = sender->locptr | mroot_mask;
  v1 = sender;
  r = loc_root;
  for (i = 0; i < TAPESTRY_NODEID_SIZE; i++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * i);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * i);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE * i + digit];
    v2->locptr = v2->locptr | mroot_mask;
    v1 = v2;
  }
}

/************************************************/
/************************************************/

void join_multicast_group(Graph *g, Vertex *loc_root,
		Vertex **recvr_array, int group_size,
		Vertex **mroot_array, int num_roots, char *root_outfile,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
                int TAPESTRY_NAMESPACE_BASE) {

  Vertex *v1, *v2, *r, *closest_mroot;
  int i, j, k, mroot_mask, distance, min_dist, mask, r_id, digit;
  Vertex **routing_table;
  long join_req_array[MAX_NUM_ROOTS];
  int closest_mroot_index;
  FILE *root_fp = fopen(root_outfile, "w");

  printf("\njoin_multicast_group:  num_roots = %d\n", num_roots + 1);
  fflush(stdout);

  for (i = 0; i < MAX_NUM_ROOTS; i++) {
    join_req_array[i] = 0;
  }

  printf("after initializing join_req_array\n");
  fflush(stdout);

  for (i = 0; i < group_size; i++) {
    v1 = recvr_array[i];
    r = loc_root;
    for (j = 0; j < TAPESTRY_NODEID_SIZE; j++) {
      routing_table = (Vertex **) v1->rtable;
      mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * j);
      tapestry_nodeID_atoi(r->nodeid, &r_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * j);
      v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j + digit];
      min_dist = INT_MAX;
      closest_mroot = NULL;
      for (k = 0; k <= num_roots; k++) {
        mroot_mask = 1 << k;
        if ((v2->locptr & mroot_mask) != 0) {
          distance = dijkstra(v2, mroot_array[k], g, NULL);
          if (distance < min_dist) {
            min_dist = distance;
            closest_mroot = mroot_array[k];
            closest_mroot_index = k;
          }
        }
      }
      if (closest_mroot != NULL) {
        (recvr_array[i])->mroot = closest_mroot;
        join_req_array[closest_mroot_index] += 1;
	printf("%s:  closest_mroot = %s, index = %d\n",
		(recvr_array[i])->name, closest_mroot->nodeid,
		closest_mroot_index);
	fflush(stdout);
        break;
      }
      v1 = v2;
    }
  }
  fwrite(join_req_array, sizeof(long), MAX_NUM_ROOTS, root_fp);
  fclose(root_fp);
}

/************************************************/
/* initialize output file names                 */
/************************************************/

void exp15_init_output_file_names(char *topology_fn, char *out_dir,
        char *hop_outfile1, char *hop_outfile2,
        char *rdp_outfile, char *stress_outfile1, char *stress_outfile2,
        char *root_outfile,
        int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
        int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
        int group_size, int num_roots, int vnn) {

  char param[MAX_FILENAME_LEN];

  //  int vnn = atoi(getenv("REXEC_MY_VNN"));

  // initialize output file names
  sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, num_roots, vnn);
  sprintf(hop_outfile1, "%s.hop.sp", param);
  sprintf(hop_outfile2, "%s.hop.tapestry", param);
  sprintf(rdp_outfile, "%s.hop.rdp", param);
  sprintf(stress_outfile1, "%s.stress.unicast", param);
  sprintf(stress_outfile2, "%s.stress.tapestry", param);

  sprintf(root_outfile, "%s.root", param);

  printf("%s\n", hop_outfile1); printf("%s\n", hop_outfile2);
  printf("%s\n", rdp_outfile);
  printf("%s\n", stress_outfile1); printf("%s\n", stress_outfile2);
  printf("%s\n", root_outfile);
}

/************************************************/
/* run experiment                               */
/* 1.  initialize output file names             */
/* 2.  clear mroot field of every vertex	*/
/* 3.  pick an additional random multicast	*/
/*     root/sender				*/
/* 4.  insert an object with the same ID at 	*/
/*     the multicast root and establish		*/
/*     location	pointers			*/ 
/* 5.  perform join operations			*/
/* 6.  calc # of hops from sender to each       */
/*     receiver via shortest path routing       */
/* 7.  calc # of hops from sender to each       */
/*     receiver via tapestry routing            */
/* 8.  compute RDP                              */
/* 9.  calc stress in unicast routing           */
/* 10. calc stress in tapestry routing          */
/************************************************/
        
void exp15_run(Graph *g, char *topology_fn, char *out_dir,
        int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
        int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
        int group_size, int num_roots, int vnn, Vertex **recvr_array,
	Vertex **mroot_array, Vertex *loc_root) {

  char hop_outfile1[MAX_FILENAME_LEN];
  char hop_outfile2[MAX_FILENAME_LEN];
  char rdp_outfile[MAX_FILENAME_LEN]; 
  char stress_outfile1[MAX_FILENAME_LEN];
  char stress_outfile2[MAX_FILENAME_LEN];
  char root_outfile[MAX_FILENAME_LEN];
  long *hop_array1, *hop_array2;
  Vertex *sender;
  
  printf("\ntopology size = %ld, tapestry size = %d, group size = %d, ",
                g->n, TAPESTRY_NAMESPACE_SIZE, group_size);
  printf("number of roots = %d\n", num_roots+1);

  exp15_init_output_file_names(topology_fn, out_dir,
                hop_outfile1, hop_outfile2, rdp_outfile,
                stress_outfile1, stress_outfile2, root_outfile,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, num_roots+1, vnn);

  // clear mroot field of every vertex
  clear_mroot_field(g);

  // pick an additional random multicast root/sender
  sender = pick_1(g);
  mroot_array[num_roots] = sender;

  // insert an object with the same ID at the multicast root
  // and establish location pointers 
  insert_object(g, loc_root, sender, num_roots,
	TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);

  // perform join operations
  join_multicast_group(g, loc_root, recvr_array, group_size,
	mroot_array, num_roots, root_outfile,
	TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);

  // calc # of hops from the sender to each member via shortest path
  hop_calc_sp_partition(g, recvr_array, group_size, hop_outfile1);

  // calc # of hops from the sender to each member via tapestry routing
  hop_calc_tapestry_partition(g, recvr_array, group_size, hop_outfile2,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);

  // compute RDP = tapestry routing delay / shortest path delay
  hop_array1 = (long *) malloc(sizeof(long) * group_size);
  hop_array2 = (long *) malloc(sizeof(long) * group_size);
  hop_read(hop_array1, group_size, hop_outfile1);
  hop_read(hop_array2, group_size, hop_outfile2);
  rdp_calc(hop_array1, hop_array2, group_size, rdp_outfile);
}

/************************************************/
/* experiment 15                                */
/* partition receivers into disjoint		*/
/* membership sets, each containing receivers	*/
/* closest to a local root in network distance	*/
/************************************************/

void exp_15() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out15_tues";
  int group_size = 4096 - (1 + MAX_NUM_ROOTS);
  int i, j, num_roots, mask;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
        TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  Vertex **recvr_array, *loc_root;
  Vertex *mroot_array[MAX_NUM_ROOTS];
  int vnn = 0;

  TAPESTRY_NODEID_SIZE = 6; 
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;          // quad digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
    
  for (i = 0; i < 4; i++) {
    g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

    clear_picked_m_field(g);

    // clear locptr field of every vertex in g
    clear_locptr_field(g);

    // pick tapestry location root from the tapestry network
    loc_root = pick_1(g);

    // pick random multicast receivers from the tapestry network
    recvr_array = (Vertex **) malloc(sizeof(Vertex*) * group_size);
    pick_n(g, recvr_array, group_size);

    for (num_roots = 0; num_roots < MAX_NUM_ROOTS; num_roots++) {
      exp15_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, num_roots, vnn, recvr_array,
		mroot_array, loc_root);

      printf("loc_root->locptr = ");
      for (j = MAX_NUM_ROOTS - 1; j >= 0; j--) {
        mask = 1 << j;
        if ((loc_root->locptr & mask) == 0)
          printf("0");
        else
          printf("1");
      }
      printf("\n");
    }
  }
}

/************************************************/
/* MAIN                                         */
/************************************************/ 
    
int main() {
   
  gb_init_rand(time(NULL));     // initialize the random number generator
  
  // experiment for 15, tree partitioning
  exp_15();

  return 0;
}               
