#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_basic.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "graph.h"
#include "exp.h"
#include "hop.h"
#include "nodeid.h"
#include "pick.h"
#include "route.h"
#include "stat.h"
#include "stress.h"
#include "util.h"

/************************************************/
/* initialize output file names                 */
/************************************************/

void exp13_14_init_output_file_names(char *topology_fn, char *out_dir,
        char *hop_outfile1, char *hop_outfile2,
        char *rdp_outfile, char *stress_outfile1, char *stress_outfile2,
        int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
        int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
        int group_size, int percentage, int vnn) {

  char param[MAX_FILENAME_LEN];

  //  int vnn = atoi(getenv("REXEC_MY_VNN"));

  // initialize output file names
  sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, percentage, vnn);
  sprintf(hop_outfile1, "%s.hop.sp", param);
  sprintf(hop_outfile2, "%s.hop.tapestry", param);
  sprintf(rdp_outfile, "%s.hop.rdp", param);
  sprintf(stress_outfile1, "%s.stress.unicast", param);
  sprintf(stress_outfile2, "%s.stress.tapestry", param);

  printf("%s\n", hop_outfile1); printf("%s\n", hop_outfile2);
  printf("%s\n", rdp_outfile);
  printf("%s\n", stress_outfile1); printf("%s\n", stress_outfile2);
}

/************************************************/
/* pick num_failed_links links randomly and	*/
/* remove them from the Graph pointed to by g	*/
/************************************************/

void inject_link_faults(Graph *g, int num_failed_links) {
  int i;
  Vertex *v, *u;
  Arc *first_arc, *a, *pred_a;

  for (i = 0; i < num_failed_links; i++) {
    first_arc = NULL;
    while (first_arc == NULL) {
      v = g->vertices + gb_unif_rand(g->n);
      first_arc = v->arcs;
    } 
    v->arcs = first_arc->next;
    u = first_arc->tip;
    pred_a = NULL;
    for (a = u->arcs; a; a = a->next) {
      if (a->tip == v) {
        if (a == u->arcs) {
          u->arcs = a->next;
        } else {
          pred_a->next = a->next; 
        }
        break;
      }
      pred_a = a;
    }
  }
}

/************************************************/
/* initialize experiment                        */
/* 1.  assign tapestry node IDs to vertices     */
/* 2.  build tapestry routing table             */
/************************************************/
    
void exp13_14_init(Graph *g, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE){

  clear_picked_t_field(g);      // clear the picked_t field of every vertex
   
  // assign tapestry nodeIDs to g's vertices
  assign_tapestry_nodeID(g, TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE,
                                TAPESTRY_NAMESPACE_SIZE);
 
  print_all_arcs(g);

  // build tapestry routing table
  build_tapestry_routing_table(g, TAPESTRY_NODEID_SIZE,
                                TAPESTRY_NAMESPACE_BASE);
  // print tapestry routing table
  print_tapestry_routing_table(g, TAPESTRY_NODEID_SIZE,
                                TAPESTRY_NAMESPACE_BASE);
}     

/************************************************/
/* run experiment                               */
/* 1.  initialize output file names             */
/* 2.  calc # of hops from sender to each       */
/*     receiver via shortest path routing       */
/* 3.  calc # of hops from sender to each       */
/*     receiver via tapestry routing            */
/* 4.  compute RDP                              */
/************************************************/
        
void exp13_14_run(Graph *g, Vertex *sender, Vertex **recvr_array,
	Graph *new_graph, Vertex *new_sender, Vertex **new_recvr_array,
	char *topology_fn, char *out_dir,
        int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
        int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
        int group_size, int percentage, int vnn) {

  char hop_outfile1[MAX_FILENAME_LEN];
  char hop_outfile2[MAX_FILENAME_LEN];
  char rdp_outfile[MAX_FILENAME_LEN]; 
  char stress_outfile1[MAX_FILENAME_LEN];
  char stress_outfile2[MAX_FILENAME_LEN];
  long *hop_array1, *hop_array2;
  
  printf("\ntopology size = %ld, tapestry size = %d, group size = %d, ",
                g->n, TAPESTRY_NAMESPACE_SIZE, group_size);
  fflush(stdout);

  exp13_14_init_output_file_names(topology_fn, out_dir,
                hop_outfile1, hop_outfile2, rdp_outfile,
                stress_outfile1, stress_outfile2,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage, vnn);
                                
  // calc # of hops from the sender to each member via shortest path
  // = -1 if original shortest path no longer exists in new_graph
  hop_calc_sp(g, sender, recvr_array,
		new_graph, new_sender, new_recvr_array,
		group_size-1, hop_outfile1);

  // calc # of hops from the sender to each member via tapestry routing
  hop_calc_tapestry(new_graph, new_sender, new_recvr_array, group_size-1,
	hop_outfile2, TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
	TAPESTRY_NAMESPACE_BASE);

  // compute RDP = tapestry routing delay / shortest path delay
  hop_array1 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_array2 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_read(hop_array1, group_size - 1, hop_outfile1);
  hop_read(hop_array2, group_size - 1, hop_outfile2);
  rdp_calc(hop_array1, hop_array2, group_size - 1, rdp_outfile);

/*
  // compute stress of a physical link = # of identifical copies of
  // a packet carried by a physical link
  clear_stress_field(g);
  stress_calc_unicast(g, sender, recvr_array, group_size-1, stress_outfile1);
  clear_stress_field(g);
  stress_calc_tapestry(g, sender, recvr_array, group_size-1, stress_outfile2,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
*/
}

/************************************************/
/* experiment 13                                */
/* inject link failures randomly		*/
/************************************************/

void exp_13() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out13";
  char gb_infile[MAX_FILENAME_LEN];
  int group_size = 256;
  int vnn = 9, i, j, k;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
        TAPESTRY_NAMESPACE_SIZE;
  Graph *g, *new_graph;
  Vertex *sender, **recvr_array, *new_sender, **new_recvr_array;
  int num_links, num_failed_links;
  double percentage = 1;		// 1%

/*
  TAPESTRY_NODEID_SIZE = 4;
  TAPESTRY_DIGIT_SIZE = 3;
  TAPESTRY_NAMESPACE_BASE = 8;          // octal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
*/

  TAPESTRY_NODEID_SIZE = 6; 
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;          // quad digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
    
  for (i = 0; i < 1; i++) {
    sprintf(gb_infile, "../data/%s", topology_fn_array[i]);
    g = restore_graph(gb_infile);	// restores graph g
    if (g == NULL)
      printf("ERROR: graph file not read from topology data file.\n");

    reset_arc_len(g);			// reset arc lengths to 1

    new_graph = restore_graph(gb_infile);
    num_links = reset_arc_len(new_graph);	// reset arc lengths to 1

    exp13_14_init(new_graph, TAPESTRY_NODEID_SIZE,
             TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

    clear_picked_m_field(new_graph);

    // pick a random sender
    new_sender = pick_1(new_graph);
    sender = g->vertices + (new_sender - new_graph->vertices);

    // pick random multicast receivers from the tapestry network
    new_recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
    pick_n(new_graph, new_recvr_array, group_size - 1);
    recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
    for (j = 0; j < group_size - 1; j++) {
      recvr_array[j] = g->vertices +
				(new_recvr_array[j] - new_graph->vertices);
    }

    k = 0;
    exp13_14_run(g, sender, recvr_array,
		new_graph, new_sender, new_recvr_array,
		topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage * k, vnn);

    num_failed_links = num_links * percentage / 100;
    for (k = 1; k <= 50; k++) {

      inject_link_faults(new_graph, num_failed_links);

      exp13_14_run(g, sender, recvr_array,
		new_graph, new_sender, new_recvr_array,
		topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage * k, vnn);
    }
  }
}

/************************************************/
/* pick num_failed_nodes nodes (not in the	*/
/* multicast group) randomly and remove them	*/
/* from the Graph pointed to by g		*/
/************************************************/

Graph *inject_node_faults(Graph *g, Vertex **new_sender,
		Vertex **new_recvr_array, int group_size,
		int num_failed_nodes,
		int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {
  Graph *new_graph;
  Vertex *v, *u, *w;
  int i, j, num_before;
  int routing_table_size = TAPESTRY_NODEID_SIZE * TAPESTRY_NAMESPACE_BASE *
                                MAX_NUM_NBRS; 
  Vertex **v_routing_table, **u_routing_table;
  Vertex *remv_array[num_failed_nodes];

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    v->ind = 1;
  }
  for (i = 0; i < num_failed_nodes; i++) {
    v = g->vertices + gb_unif_rand(g->n);
    while (v->ind == 0 || v->picked_m == 1) {
      v = g->vertices + gb_unif_rand(g->n);
    }
    v->ind = 0;
    remv_array[i] = v; 
    printf("remv_array[%d] = %s\n", i, v->name);
  }
  new_graph = induced(g, NULL, 0, 0, 0);

  reset_arc_len(new_graph);		// reset arc lengths to 1
  clear_picked_t_field(new_graph);      // clear the picked_t field
  clear_picked_m_field(new_graph);	// clear the picked_m field

  for (v = g->vertices, u = new_graph->vertices; v < g->vertices + g->n &&
			u < new_graph->vertices + new_graph->n; v++, u++) {
    while (v->ind == 0)
      v++;
    u->picked_t = v->picked_t;
    u->picked_m = v->picked_m;
    u->nodeid = v->nodeid;

    if (u->picked_t == 1) {
      v_routing_table = (Vertex **) v->rtable;
      u_routing_table = (Vertex **)malloc(sizeof(Vertex*)*routing_table_size);
      for (i = 0; i < routing_table_size; i++) {
        w = v_routing_table[i];
        num_before = 0;
        for (j = 0; j < num_failed_nodes; j++) {
          if (remv_array[j] < w)
            num_before++;
          else if (remv_array[j] == w)
            break;
        }
        if (remv_array[j] == w)
          u_routing_table[i] = NULL;
        else
          u_routing_table[i] = new_graph->vertices +
				(w - g->vertices - num_before);
      }
      u->rtable = (long) u_routing_table;
    }
  }

  // update new_sender, new_recvr_array to point to vertices in new_graph
  num_before = 0;
  for (j = 0; j < num_failed_nodes; j++) {
    if (remv_array[j] < *new_sender)
      num_before++;
    else if (remv_array[j] == *new_sender)
      break;
  }
  if (remv_array[j] == *new_sender) {
    printf("ERROR:  inject_node_faults removed sender!\n");
    exit(1);
  }
  else
    *new_sender = new_graph->vertices +
				(*new_sender - g->vertices - num_before);

  for (i = 0; i < group_size - 1; i++) {
    num_before = 0;
    for (j = 0; j < num_failed_nodes; j++) {
      if (remv_array[j] < new_recvr_array[i])
        num_before++;
      else if (remv_array[j] == new_recvr_array[i])
        break;
    }
    if (remv_array[j] == new_recvr_array[i]) {
      printf("ERROR:  inject_node_faults removed receiver %d!\n", i);
      exit(1);
    }
    else
      new_recvr_array[i] = new_graph->vertices +
			(new_recvr_array[i] - g->vertices - num_before);
  }
  return new_graph;
}

/************************************************/
/* experiment 14                                */
/* inject node failures randomly                */
/************************************************/

void exp_14() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out14_try";
  char gb_infile[MAX_FILENAME_LEN];
  int group_size = 256;
  int vnn = 9, i, j, k;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
        TAPESTRY_NAMESPACE_SIZE;
  Graph *g, *new_graph;
  Vertex *sender, **recvr_array, *new_sender, **new_recvr_array;
  int num_failed_nodes;
  double percentage = 1;                // 1%

  TAPESTRY_NODEID_SIZE = 6;
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;          // quad digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
 
  for (i = 1; i < 2; i++) {
    sprintf(gb_infile, "../data/%s", topology_fn_array[i]);
    g = restore_graph(gb_infile);       // restores graph g
    if (g == NULL)
      printf("ERROR: graph file not read from topology data file.\n");

    reset_arc_len(g);			// reset arc lengths to 1

    new_graph = restore_graph(gb_infile);
    reset_arc_len(new_graph);		// reset arc lengths to 1

    exp13_14_init(new_graph, TAPESTRY_NODEID_SIZE,
             TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

    clear_picked_m_field(new_graph);

    // pick a random sender
    new_sender = pick_1(new_graph);
    sender = g->vertices + (new_sender - new_graph->vertices);

    // pick random multicast receivers from the tapestry network
    new_recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
    pick_n(new_graph, new_recvr_array, group_size - 1);
    recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
    for (j = 0; j < group_size - 1; j++) {
      recvr_array[j] = g->vertices +
                                (new_recvr_array[j] - new_graph->vertices);
    }

    k = 0;
    exp13_14_run(g, sender, recvr_array,
                new_graph, new_sender, new_recvr_array,
                topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage * k, vnn);

    num_failed_nodes = g->n * percentage / 100;
    for (k = 1; k <= 50; k++) {
      new_graph = inject_node_faults(new_graph, &new_sender, new_recvr_array,
		group_size, num_failed_nodes,
                TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);

      exp13_14_run(g, sender, recvr_array,
                new_graph, new_sender, new_recvr_array,
                topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, percentage * k, vnn);
    }
  }
}

/************************************************/
/* MAIN                                         */
/************************************************/ 
    
int main() {
   
  gb_init_rand(time(NULL));     // initialize the random number generator
  
/*
  // experiment for 13, random link failures
  exp_13();
*/

  // experiment for 14, random node failures
  exp_14();

  return 0;
}               
