#include <limits.h>
#include <gb_graph.h>
#include <gb_dijk.h>

#include "exp.h"
#include "graph.h"
#include "nodeid.h"
#include "stress.h"

/************************************************/
/* clear the stress field of each arc of        */
/* the Graph pointed to by g                    */
/************************************************/
  
void clear_stress_field(Graph *g) {
  Vertex *v;
  Arc *a;

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    for (a = v->arcs; a; a = a->next) {
      a->stress = 0;
    }
  }
}
                
/************************************************/
/* accumulate stress of every physical link     */
/* on the dijkstra shortest path between        */
/* the two input vertices v1 and v2             */
/************************************************/
    
void accumulate_physical_link_stress(Graph *g, Vertex *v1, Vertex *v2) {
  int i, distance;
  Vertex *u1, *u2;
  Arc *a;

  dijkstra(v1, v2, g, NULL);
  distance = v2->dist;
  printf("%s(%d): ", v2->nodeid, distance);
  
  // for each edge on the dijkstra shortest path, accumulate its stress
  for (i = 0; i < distance; i++) {
    if (strcmp(v2->name, v2->backlink->name) < 0) {
      u1 =  v2;
      u2 = v2->backlink;
    } else if (strcmp(v2->name, v2->backlink->name) > 0) {
      u1 = v2->backlink;
      u2 = v2;
    } else {                          // should not happen!
      printf("ERROR:  two vertices have the same name %s!\n", v2->name);
    }
    for (a = u1->arcs; a; a = a->next) {
      if (a->tip == u2) {
        a->stress += 1;
        break;
      }
    }
    printf("%s ", v2->backlink->nodeid);
    v2 = v2->backlink;
  }
  printf("\n");
}

/************************************************/
/* calculate stress of physical links from      */
/* the sender to each receiver in the           */
/* recvr_array via unicast shortest path        */
/* routing, and write to stress_outfile         */
/************************************************/

void stress_calc_unicast(Graph *g, Vertex *sender, Vertex **recvr_array,
                        int n, char *stress_outfile) {
  int i;
  FILE *stress_fp = fopen(stress_outfile, "w");
  Vertex *v = NULL;
  Arc *a = NULL;
      
  printf("\nstress_calc_unicast:  sender %s (%s)\n",
					sender->name, sender->nodeid);
  for (i = 0; i < n; i++) {
    v = recvr_array[i];
    accumulate_physical_link_stress(g, sender, v);
  }
  printf("\n");

  // write physical link stress to file in ascii
  // first column is the link, represented as vertex1<->vertex2
  // second column is the physical stress of that edge
  for (v = g->vertices; v < g->vertices + g->n; v++) {
    for (a = v->arcs; a; a = a->next) {
      if (a->stress > 0) {
        fprintf(stress_fp, "%s<->%s %ld\n", v->name, a->tip->name, a->stress);
      }
    }
  }
  fclose(stress_fp);
}

/************************************************/
/* simulate tapestry multicast routing          */
/************************************************/

void stress_calc_tapestry_recurse(Graph *g, Vertex *sender,
        Vertex **recvr_array, int n, int level, int TAPESTRY_NODEID_SIZE,
        int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  Vertex *v, *s, **routing_table, **r_array;
  int i, j, v_id, v_mask, v_suffix, count;

  if (level > TAPESTRY_NODEID_SIZE)
    return;

  v_mask = (TAPESTRY_NAMESPACE_BASE-1) << (TAPESTRY_DIGIT_SIZE*(level-1));

  for (i = 0; i < TAPESTRY_NAMESPACE_BASE; i++) {
    count = 0;
    for (j = 0; j < n; j++) {
      v = recvr_array[j];
      tapestry_nodeID_atoi(v->nodeid, &v_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      v_suffix = (v_id & v_mask) >> (TAPESTRY_DIGIT_SIZE * (level - 1));
      if (v_suffix == i)
        count++;
    }
    if (count > 0) {
      printf("at %s: level %d (%d): ", sender->nodeid, level, i);
      r_array = (Vertex **) malloc(sizeof(Vertex *) * count);
      count = 0;
      for (j = 0; j < n; j++) {
        v = recvr_array[j];
        tapestry_nodeID_atoi(v->nodeid, &v_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
        v_suffix = (v_id & v_mask) >> (TAPESTRY_DIGIT_SIZE * (level - 1));
        if (v_suffix == i) {
          printf("%s ", v->nodeid);
          r_array[count++] = v;
        }
      }
      printf("\n");
      routing_table = (Vertex **) sender->rtable;
      s = routing_table[TAPESTRY_NAMESPACE_BASE * (level-1) + i];
      accumulate_physical_link_stress(g, sender, s);
      stress_calc_tapestry_recurse(g, s, r_array, count, level+1,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
    }
  }
}

/************************************************/
/* calculate stress of physical links from      */
/* the sender to each receiver in the           */
/* recvr_array via tapestry multicast routing,  */
/* and write to stress_outfile                  */
/************************************************/

void stress_calc_tapestry(Graph *g, Vertex *sender, Vertex **recvr_array,
                int n, char *stress_outfile, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  FILE *stress_fp = fopen(stress_outfile, "w");
  Vertex *v = NULL;
  Arc *a = NULL;

  printf("stress_calc_tapestry:  sender %s (%s)\n",
					sender->name, sender->nodeid);
  stress_calc_tapestry_recurse(g, sender, recvr_array, n, 1,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);

  // write physical link stress to file in ascii
  // first column is the link, represented as vertex1<->vertex2
  // second column is the physical stress of that edge
  for (v = g->vertices; v < g->vertices + g->n; v++) {
    for (a = v->arcs; a; a = a->next) {
      if (a->stress > 0) {
        fprintf(stress_fp, "%s<->%s %ld\n", v->name, a->tip->name, a->stress);
      }
    }
  }
  fclose(stress_fp);
}
